import os
import pkg_resources
from .files import read_file

import logging

logger = logging.getLogger(__name__)

__version__ = "0.0.0"
version_filename = "VERSION"

wm_path: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))
wm_version_file = f"{wm_path}/{version_filename}"

if pkg_resources.resource_exists(__name__, version_filename):
    __version__ = pkg_resources.resource_string(__name__, version_filename).decode("utf-8").strip()
elif os.path.exists(wm_version_file):
    __version__ = read_file(wm_version_file)
else:
    logger.warning("No version found")
    pass



#!/usr/bin/env python3
"""CSS Minifier functions for CSS-HTML-JS-Minify."""


import re
import itertools
import datetime

from .variables import EXTENDED_NAMED_COLORS, CSS_PROPS_TEXT


__all__ = ("css_minify", "condense_semicolons")


def _compile_props(props_text, grouped=False):
    """Take a list of props and prepare them."""
    props, prefixes = [], "-webkit-,-khtml-,-epub-,-moz-,-ms-,-o-,".split(",")
    for propline in props_text.strip().lower().splitlines():
        props += [pre + pro for pro in propline.split(" ") for pre in prefixes]
    props = filter(lambda line: not line.startswith("#"), props)
    if not grouped:
        props = list(filter(None, props))
        return props, [0] * len(props)
    final_props, groups, g_id = [], [], 0
    for prop in props:
        if prop.strip():
            final_props.append(prop)
            groups.append(g_id)
        else:
            g_id += 1
    return final_props, groups


def _prioritify(line_of_css, css_props_text_as_list):
    """Return args priority, priority is integer and smaller means higher."""
    sorted_css_properties, groups_by_alphabetic_order = css_props_text_as_list
    priority_integer, group_integer = 9999, 0
    for css_property in sorted_css_properties:
        if css_property.lower() == line_of_css.split(":")[0].lower().strip():
            priority_integer = sorted_css_properties.index(css_property)
            group_integer = groups_by_alphabetic_order[priority_integer]
            break
    return priority_integer, group_integer


def _props_grouper(props, pgs):
    """Return groups for properties."""
    if not props:
        return props
    # props = sorted([
    # _ if _.strip().endswith(";")
    # and not _.strip().endswith("*/") and not _.strip().endswith("/*")
    # else _.rstrip() + ";\n" for _ in props])
    props_pg = zip(map(lambda prop: _prioritify(prop, pgs), props), props)
    props_pg = sorted(props_pg, key=lambda item: item[0][1])
    props_by_groups = map(lambda item: list(item[1]), itertools.groupby(props_pg, key=lambda item: item[0][1]))
    props_by_groups = map(lambda item: sorted(item, key=lambda item: item[0][0]), props_by_groups)
    props = []
    for group in props_by_groups:
        group = map(lambda item: item[1], group)
        props += group
        props += ["\n"]
    props.pop()
    return props


def sort_properties(css_unsorted_string):
    """CSS Property Sorter Function.

    This function will read buffer argument, split it to a list by lines,
    sort it by defined rule, and return sorted buffer if it's CSS property.
    This function depends on '_prioritify' function.
    """
    css_pgs = _compile_props(CSS_PROPS_TEXT, grouped=False)  # Do Not Group.
    pattern = re.compile(r"(.*?{\r?\n?)(.*?)(}.*?)|(.*)", re.DOTALL + re.MULTILINE)
    matched_patterns = pattern.findall(css_unsorted_string)
    sorted_patterns, sorted_buffer = [], css_unsorted_string
    re_prop = re.compile(r"((?:.*?)(?:;)(?:.*?\n)|(?:.*))", re.DOTALL + re.MULTILINE)
    if len(matched_patterns) != 0:
        for matched_groups in matched_patterns:
            sorted_patterns += matched_groups[0].splitlines(True)
            props = map(lambda line: line.lstrip("\n"), re_prop.findall(matched_groups[1]))
            props = list(filter(lambda line: line.strip("\n "), props))
            props = _props_grouper(props, css_pgs)
            sorted_patterns += props
            sorted_patterns += matched_groups[2].splitlines(True)
            sorted_patterns += matched_groups[3].splitlines(True)
        sorted_buffer = "".join(sorted_patterns)
    return sorted_buffer


def remove_comments(css):
    """Remove all CSS comment blocks."""
    iemac, preserve = False, False
    comment_start = css.find("/*")
    while comment_start >= 0:  # Preserve comments that look like `/*!...*/`.
        # Slicing is used to make sure we dont get an IndexError.
        preserve = css[comment_start + 2 : comment_start + 3] == "!"
        comment_end = css.find("*/", comment_start + 2)
        if comment_end < 0:
            if not preserve:
                css = css[:comment_start]
                break
        elif comment_end >= (comment_start + 2):
            if css[comment_end - 1] == "\\":
                # This is an IE Mac-specific comment; leave this one and the
                # following one alone.
                comment_start = comment_end + 2
                iemac = True
            elif iemac:
                comment_start = comment_end + 2
                iemac = False
            elif not preserve:
                css = css[:comment_start] + css[comment_end + 2 :]
            else:
                comment_start = comment_end + 2
        comment_start = css.find("/*", comment_start)
    return css


def remove_unnecessary_whitespace(css):
    """Remove unnecessary whitespace characters."""

    def pseudoclasscolon(css):
        """Prevent 'p :link' from becoming 'p:link'.

        Translates 'p :link' into 'p ___PSEUDOCLASSCOLON___link'.
        This is translated back again later.
        """
        regex = re.compile(r"(^|\})(([^\{\:])+\:)+([^\{]*\{)")
        match = regex.search(css)
        while match:
            css = "".join([css[: match.start()], match.group().replace(":", "___PSEUDOCLASSCOLON___"), css[match.end() :]])
            match = regex.search(css)
        return css

    css = pseudoclasscolon(css)
    # Remove spaces from before things.
    css = re.sub(r"\s+([!{};:>\(\)\],])", r"\1", css)
    # If there is a `@charset`, then only allow one, and move to beginning.
    css = re.sub(r"^(.*)(@charset \"[^\"]*\";)", r"\2\1", css)
    css = re.sub(r"^(\s*@charset [^;]+;\s*)+", r"\1", css)
    # Put the space back in for a few cases, such as `@media screen` and
    # `(-webkit-min-device-pixel-ratio:0)`.
    css = re.sub(r"\band\(", "and (", css)
    # Put the colons back.
    css = css.replace("___PSEUDOCLASSCOLON___", ":")
    # Remove spaces from after things.
    css = re.sub(r"([!{}:;>\(\[,])\s+", r"\1", css)
    return css


def remove_unnecessary_semicolons(css):
    """Remove unnecessary semicolons."""
    return re.sub(r";+\}", "}", css)


def remove_empty_rules(css):
    """Remove empty rules."""
    return re.sub(r"[^\}\{]+\{\}", "", css)


def normalize_rgb_colors_to_hex(css):
    """Convert `rgb(51,102,153)` to `#336699`."""
    regex = re.compile(r"rgb\s*\(\s*([0-9,\s]+)\s*\)")
    match = regex.search(css)
    while match:
        colors = map(lambda s: s.strip(), match.group(1).split(","))
        hexcolor = "#%.2x%.2x%.2x" % tuple(map(int, colors))
        css = css.replace(match.group(), hexcolor)
        match = regex.search(css)
    return css


def condense_zero_units(css):
    """Replace `0(px, em, %, etc)` with `0`."""
    return re.sub(r"([\s:])(0)(px|em|%|in|q|ch|cm|mm|pc|pt|ex|rem|s|ms|" r"deg|grad|rad|turn|vw|vh|vmin|vmax|fr)", r"\1\2", css)


def condense_multidimensional_zeros(css):
    """Replace `:0 0 0 0;`, `:0 0 0;` etc. with `:0;`."""
    return css.replace(":0 0 0 0;", ":0;").replace(":0 0 0;", ":0;").replace(":0 0;", ":0;").replace("background-position:0;", "background-position:0 0;").replace("transform-origin:0;", "transform-origin:0 0;")


def condense_floating_points(css):
    """Replace `0.6` with `.6` where possible."""
    return re.sub(r"(:|\s)0+\.(\d+)", r"\1.\2", css)


def condense_hex_colors(css):
    """Shorten colors from #AABBCC to #ABC where possible."""
    regex = re.compile(r"""([^\"'=\s])(\s*)#([0-9a-f])([0-9a-f])([0-9a-f])""" r"""([0-9a-f])([0-9a-f])([0-9a-f])""", re.I | re.S)
    match = regex.search(css)
    while match:
        first = match.group(3) + match.group(5) + match.group(7)
        second = match.group(4) + match.group(6) + match.group(8)
        if first.lower() == second.lower():
            css = css.replace(match.group(), match.group(1) + match.group(2) + "#" + first)
            match = regex.search(css, match.end() - 3)
        else:
            match = regex.search(css, match.end())
    return css


def condense_whitespace(css):
    """Condense multiple adjacent whitespace characters into one."""
    return re.sub(r"\s+", " ", css)


def condense_semicolons(css):
    """Condense multiple adjacent semicolon characters into one."""
    return re.sub(r";;+", ";", css)


def wrap_css_lines(css, line_length=80):
    """Wrap the lines of the given CSS to an approximate length."""
    lines, line_start = [], 0
    for i, char in enumerate(css):
        # Its safe to break after } characters.
        if char == "}" and (i - line_start >= line_length):
            lines.append(css[line_start : i + 1])
            line_start = i + 1
    if line_start < len(css):
        lines.append(css[line_start:])
    return "\n".join(lines)


def condense_font_weight(css):
    """Condense multiple font weights into shorter integer equals."""
    return css.replace("font-weight:normal;", "font-weight:400;").replace("font-weight:bold;", "font-weight:700;")


def condense_std_named_colors(css):
    """Condense named color values to shorter replacement using HEX."""
    for color_name, color_hexa in iter(tuple({":aqua;": ":#0ff;", ":blue;": ":#00f;", ":fuchsia;": ":#f0f;", ":yellow;": ":#ff0;"}.items())):
        css = css.replace(color_name, color_hexa)
    return css


def condense_xtra_named_colors(css):
    """Condense named color values to shorter replacement using HEX."""
    for k, v in iter(tuple(EXTENDED_NAMED_COLORS.items())):
        same_color_but_rgb = "rgb({0},{1},{2})".format(v[0], v[1], v[2])
        if len(k) > len(same_color_but_rgb):
            css = css.replace(k, same_color_but_rgb)
    return css


def remove_url_quotes(css):
    """Fix for url() does not need quotes."""
    return re.sub(r'url\((["\'])([^)]*)\1\)', r"url(\2)", css)


def condense_border_none(css):
    """Condense border:none; to border:0;."""
    return css.replace("border:none;", "border:0;")


def add_encoding(css):
    """Add @charset 'UTF-8'; if missing."""
    return '@charset "utf-8";' + css if "@charset" not in css.lower() else css


def restore_needed_space(css):
    """Fix CSS for some specific cases where a white space is needed."""
    return css.replace("!important", " !important").replace("@media(", "@media (").replace("data:image/jpeg;base64,", "data:image/jpg;base64,").rstrip("\n;")  # !important  # media queries # jpeg > jpg


def unquote_selectors(css):
    """Fix CSS for some specific selectors where Quotes is not needed."""
    return re.compile('([a-zA-Z]+)="([a-zA-Z0-9-_\.]+)"]').sub(r"\1=\2]", css)


def add_timestamp(css):
    return f"{css}\n/* {format(datetime.datetime.now().isoformat()[:-7].lower())} */\n"


def css_minify(css, settings: dict):
    """Minify CSS main function."""
    css = remove_comments(css) if not settings.get("comments", False) else css
    css = sort_properties(css) if settings.get("sort") else css
    css = unquote_selectors(css)
    css = condense_whitespace(css)
    css = remove_url_quotes(css)
    css = condense_xtra_named_colors(css)
    css = condense_font_weight(css)
    css = remove_unnecessary_whitespace(css)
    css = condense_std_named_colors(css)
    css = remove_unnecessary_semicolons(css)
    css = condense_zero_units(css)
    css = condense_multidimensional_zeros(css)
    css = condense_floating_points(css)
    css = normalize_rgb_colors_to_hex(css)
    css = condense_hex_colors(css)
    css = condense_border_none(css)
    css = wrap_css_lines(css, 80) if settings.get("wrap", False) else css
    css = condense_semicolons(css)
    css = add_encoding(css) if settings.get("prefix", False) else css
    css = restore_needed_space(css)
    css = add_timestamp(css) if settings.get("timestamp", False) else css
    return css.strip()



# https://engineeringblog.yelp.com/2017/06/making-photos-smaller.html


def jpeg_minify(jpeg, settings: dict):
    """Minify JPEG main function."""
    # TODO: Implement this
    # print(f"JPEG WAS CALLED FOR {len(jpeg)}!!!!! #############################################")
    return jpeg



from .html import html_minify
from .css import css_minify
from .js import js_minify
from .sass import sass_minify
from .png import png_minify
from .jpeg import jpeg_minify
from .svg import svg_minify

__all__ = [html_minify, css_minify, js_minify, sass_minify, png_minify, jpeg_minify, svg_minify]



from .css import css_minify

import logging

logger = logging.getLogger(__name__)

have_sass = False

try:
    import sass

    have_sass = True
except Exception:
    logger.warning("Could not import sass", exc_info=True)
    logger.warning("NOTE: To install sass, please remember to use  use pip install libsass instead of pip install sass")


def sass_minify(css, settings: dict):
    """Minify SASS main function."""
    if have_sass:
        css = sass.compile(string=css)
    css = css_minify(css=css, settings=settings)
    return css



# https://engineeringblog.yelp.com/2017/06/making-photos-smaller.html


# import pngquant


def png_minify(png, settings: dict):
    """Minify PNG main function."""
    #    pngquant.config()
    #    png=quant_data(self, data=None, dst=None, ndeep=None, ndigits=None, delete=True):
    # TODO: Implement this
    return png



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Variables for CSS processing for CSS-HTML-JS-Minify."""


__all__ = ("EXTENDED_NAMED_COLORS", "CSS_PROPS_TEXT")


# 'Color Name String': (R, G, B)
EXTENDED_NAMED_COLORS = {
    "azure": (240, 255, 255),
    "beige": (245, 245, 220),
    "bisque": (255, 228, 196),
    "blanchedalmond": (255, 235, 205),
    "brown": (165, 42, 42),
    "burlywood": (222, 184, 135),
    "chartreuse": (127, 255, 0),
    "chocolate": (210, 105, 30),
    "coral": (255, 127, 80),
    "cornsilk": (255, 248, 220),
    "crimson": (220, 20, 60),
    "cyan": (0, 255, 255),
    "darkcyan": (0, 139, 139),
    "darkgoldenrod": (184, 134, 11),
    "darkgray": (169, 169, 169),
    "darkgreen": (0, 100, 0),
    "darkgrey": (169, 169, 169),
    "darkkhaki": (189, 183, 107),
    "darkmagenta": (139, 0, 139),
    "darkolivegreen": (85, 107, 47),
    "darkorange": (255, 140, 0),
    "darkorchid": (153, 50, 204),
    "darkred": (139, 0, 0),
    "darksalmon": (233, 150, 122),
    "darkseagreen": (143, 188, 143),
    "darkslategray": (47, 79, 79),
    "darkslategrey": (47, 79, 79),
    "darkturquoise": (0, 206, 209),
    "darkviolet": (148, 0, 211),
    "deeppink": (255, 20, 147),
    "dimgray": (105, 105, 105),
    "dimgrey": (105, 105, 105),
    "firebrick": (178, 34, 34),
    "forestgreen": (34, 139, 34),
    "gainsboro": (220, 220, 220),
    "gold": (255, 215, 0),
    "goldenrod": (218, 165, 32),
    "gray": (128, 128, 128),
    "green": (0, 128, 0),
    "grey": (128, 128, 128),
    "honeydew": (240, 255, 240),
    "hotpink": (255, 105, 180),
    "indianred": (205, 92, 92),
    "indigo": (75, 0, 130),
    "ivory": (255, 255, 240),
    "khaki": (240, 230, 140),
    "lavender": (230, 230, 250),
    "lavenderblush": (255, 240, 245),
    "lawngreen": (124, 252, 0),
    "lemonchiffon": (255, 250, 205),
    "lightcoral": (240, 128, 128),
    "lightcyan": (224, 255, 255),
    "lightgray": (211, 211, 211),
    "lightgreen": (144, 238, 144),
    "lightgrey": (211, 211, 211),
    "lightpink": (255, 182, 193),
    "lightsalmon": (255, 160, 122),
    "lightseagreen": (32, 178, 170),
    "lightslategray": (119, 136, 153),
    "lightslategrey": (119, 136, 153),
    "lime": (0, 255, 0),
    "limegreen": (50, 205, 50),
    "linen": (250, 240, 230),
    "magenta": (255, 0, 255),
    "maroon": (128, 0, 0),
    "mediumorchid": (186, 85, 211),
    "mediumpurple": (147, 112, 219),
    "mediumseagreen": (60, 179, 113),
    "mediumspringgreen": (0, 250, 154),
    "mediumturquoise": (72, 209, 204),
    "mediumvioletred": (199, 21, 133),
    "mintcream": (245, 255, 250),
    "mistyrose": (255, 228, 225),
    "moccasin": (255, 228, 181),
    "navy": (0, 0, 128),
    "oldlace": (253, 245, 230),
    "olive": (128, 128, 0),
    "olivedrab": (107, 142, 35),
    "orange": (255, 165, 0),
    "orangered": (255, 69, 0),
    "orchid": (218, 112, 214),
    "palegoldenrod": (238, 232, 170),
    "palegreen": (152, 251, 152),
    "paleturquoise": (175, 238, 238),
    "palevioletred": (219, 112, 147),
    "papayawhip": (255, 239, 213),
    "peachpuff": (255, 218, 185),
    "peru": (205, 133, 63),
    "pink": (255, 192, 203),
    "plum": (221, 160, 221),
    "purple": (128, 0, 128),
    "rosybrown": (188, 143, 143),
    "saddlebrown": (139, 69, 19),
    "salmon": (250, 128, 114),
    "sandybrown": (244, 164, 96),
    "seagreen": (46, 139, 87),
    "seashell": (255, 245, 238),
    "sienna": (160, 82, 45),
    "silver": (192, 192, 192),
    "slategray": (112, 128, 144),
    "slategrey": (112, 128, 144),
    "snow": (255, 250, 250),
    "springgreen": (0, 255, 127),
    "teal": (0, 128, 128),
    "thistle": (216, 191, 216),
    "tomato": (255, 99, 71),
    "turquoise": (64, 224, 208),
    "violet": (238, 130, 238),
    "wheat": (245, 222, 179),
}


# Do Not compact this string, new lines are used to Group up stuff.
CSS_PROPS_TEXT = """

alignment-adjust alignment-baseline animation animation-delay
animation-direction animation-duration animation-iteration-count
animation-name animation-play-state animation-timing-function appearance
azimuth

backface-visibility background background-blend-mode background-attachment
background-clip background-color background-image background-origin
background-position background-position-block background-position-inline
background-position-x background-position-y background-repeat background-size
baseline-shift bikeshedding bookmark-label bookmark-level bookmark-state
bookmark-target border border-bottom border-bottom-color
border-bottom-left-radius border-bottom-parts border-bottom-right-radius
border-bottom-style border-bottom-width border-clip border-clip-top
border-clip-right border-clip-bottom border-clip-left border-collapse
border-color border-corner-shape border-image border-image-outset
border-image-repeat border-image-slice border-image-source border-image-width
border-left border-left-color border-left-style border-left-parts
border-left-width border-limit border-parts border-radius border-right
border-right-color border-right-style border-right-width border-right-parts
border-spacing border-style border-top border-top-color border-top-left-radius
border-top-parts border-top-right-radius border-top-style border-top-width
border-width bottom box-decoration-break box-shadow box-sizing

caption-side clear clip color column-count column-fill column-gap column-rule
column-rule-color column-rule-style column-rule-width column-span column-width
columns content counter-increment counter-reset corners corner-shape
cue cue-after cue-before cursor

direction display drop-initial-after-adjust drop-initial-after-align
drop-initial-before-adjust drop-initial-before-align drop-initial-size
drop-initial-value

elevation empty-cells

flex flex-basis flex-direction flex-flow flex-grow flex-shrink flex-wrap fit
fit-position float font font-family font-size font-size-adjust font-stretch
font-style font-variant font-weight

grid-columns grid-rows

justify-content

hanging-punctuation height hyphenate-character hyphenate-resource hyphens

icon image-orientation image-resolution inline-box-align

left letter-spacing line-height line-stacking line-stacking-ruby
line-stacking-shift line-stacking-strategy linear-gradient list-style
list-style-image list-style-position list-style-type

margin margin-bottom margin-left margin-right margin-top marquee-direction
marquee-loop marquee-speed marquee-style max-height max-width min-height
min-width

nav-index

opacity orphans outline outline-color outline-offset outline-style
outline-width overflow overflow-style overflow-x overflow-y

padding padding-bottom padding-left padding-right padding-top page
page-break-after page-break-before page-break-inside pause pause-after
pause-before perspective perspective-origin pitch pitch-range play-during
position presentation-level

quotes

resize rest rest-after rest-before richness right rotation rotation-point
ruby-align ruby-overhang ruby-position ruby-span

size speak speak-header speak-numeral speak-punctuation speech-rate src
stress string-set

table-layout target target-name target-new target-position text-align
text-align-last text-decoration text-emphasis text-indent text-justify
text-outline text-shadow text-transform text-wrap top transform
transform-origin transition transition-delay transition-duration
transition-property transition-timing-function

unicode-bidi unicode-range

vertical-align visibility voice-balance voice-duration voice-family
voice-pitch voice-range voice-rate voice-stress voice-volume volume

white-space widows width word-break word-spacing word-wrap

z-index

"""



import os
from time import sleep

import logging

logger = logging.getLogger(__name__)

# fmt: off
file_endings = {
    ".sass": "sass",
    ".css": "css",
    ".js": "js",
    ".html": "html",
    ".htm": "html",
    ".tpl": "html",
    ".svg": "svg",
    ".png": "png",
    ".jpeg": "jpeg",
    ".jpg": "jpeg"
}
# fmt: on


def generate_file_list(folder: str, target: tuple = (), omit: tuple = (), allow_hidden: bool = False, topdown: bool = True, onerror: object = None, followlinks: bool = False) -> tuple:
    oswalk = os.walk(folder, topdown=topdown, onerror=onerror, followlinks=followlinks)
    return [
        os.path.abspath(os.path.join(root, file))
        for root, dirs, files in oswalk
        for file in files
        if not file.startswith(() if allow_hidden else ".") and not file.endswith(omit)
        #        and (f.endswith(target) if target else True)
    ]


def path_is_in_path(path_a, path_b):
    if not path_a or not path_b:
        return False
    return os.path.abspath(path_a) == os.path.commonpath([os.path.abspath(path_a), os.path.abspath(path_b)])


def watch_files(file_path, settings: dict):
    """Process multiple CSS, JS, HTML files with multiprocessing."""
    logger.info(f"Process {os.getpid()} is Compressing {file_path}.")
    if settings.get("watch", False):
        previous = int(os.stat(file_path).st_mtime)
        logger.info(f"Process {os.getpid()} is Watching {file_path}.")
        # logger.info(f'Total Maximum CPUs used: ~{cpu_count()} Cores.')
        #        pool = Pool(cpu_count())  # Multiprocessing Async
        #        pool.map_async(partial(
        #                watch_files, settings=args),
        #           list_of_files)
        #      pool.close()
        #     pool.join()

        while True:
            actual = int(os.stat(file_path).st_mtime)
            if previous == actual:
                sleep(60)
            else:
                previous = actual
                logger.info(f"Modification detected on {file_path}.")
                process_single_file(file_path=file_path, settings=settings)
    else:
        process_single_file(file_path=file_path, settings=settings)


def determine_file_extension(file_path: str):
    if not file_path:
        return None
    index = file_path.rfind(".")
    if index < 0:
        return None
    file_ending = file_path[index:]
    if not file_ending:
        return None
    return file_endings.get(file_ending, None)


def read_file(fname, strip=True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    data = ""
    if os.path.exists(fn):
        with open(fn) as f:
            data = f.read()
            data = data.strip() if strip else data
            # logger.info(f"Got data '{data}' from '{fn}'")
    else:
        logger.error(f"Could not find file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return data



#!/usr/bin/env python

# To generate DEB package from Python Package:
# sudo pip3 install stdeb
# python3 setup.py --verbose --command-packages=stdeb.command bdist_deb
#
#
# To generate RPM package from Python Package:
# sudo apt-get install rpm
# python3 setup.py bdist_rpm --verbose --fix-python --binary-only
#
#
# To generate EXE MS Windows from Python Package (from MS Windows only):
# python3 setup.py bdist_wininst --verbose
#
#
# To generate PKGBUILD ArchLinux from Python Package (from PyPI only):
# sudo pip3 install git+https://github.com/bluepeppers/pip2arch.git
# pip2arch.py PackageNameHere
#
#
# To Upload to PyPI by executing:
# python3 setup.py register
# python3 setup.py bdist_egg sdist --formats=zip upload --sign


import os
import pprint
from pathlib import Path
from shutil import copytree, which
from setuptools import setup, find_packages, Command
from tempfile import TemporaryDirectory
from zipapp import create_archive

import logging

logger = logging.getLogger(__name__)

setup_requirements = ["pytest-runner", "setuptools_scm"]

from web_minify import version_filename, __version__ as web_minify_version


def read_file(fname, strip=True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    data = ""
    if os.path.exists(fn):
        with open(fn) as f:
            data = f.read()
            data = data.strip() if strip else data
            # logger.info(f"Got data '{data}' from '{fn}'")
    else:
        logger.error(f"Could not find file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return data


def remove_comment(line, sep="#"):
    i = line.find(sep)
    if i >= 0:
        line = line[:i]
    return line.strip()


def read_requirements_file(fname: str):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    print(f"Reading requirements from {fn}")
    lines = []
    if os.path.exists(fn):
        with open(fn) as f:
            for r in f.readlines():
                r = r.strip()
                if len(r) < 1:
                    continue
                r = remove_comment(r)
                if len(r) < 1:
                    continue
                lines.append(r)
    else:
        logger.error(f"Could not find requirements file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return lines


##############################################################################
# Dont touch below


class ZipApp(Command):
    description, user_options = "Creates a ZipApp.", []

    def initialize_options(self):
        pass  # Dont needed, but required.

    def finalize_options(self):
        pass  # Dont needed, but required.

    def run(self):
        with TemporaryDirectory() as tmpdir:
            copytree(".", Path(tmpdir) / "web-minify")
            (Path(tmpdir) / "__main__.py").write_text("import runpy\nrunpy.run_module('web-minify')")
            create_archive(tmpdir, "web-minify.pyz", which("python3"))


##############################################################################

package = {
    "name": "web-minify",
    "version": web_minify_version,
    "author": "Lennart Rolland",
    "author_email": "lennart@firtkissdesigns.com",
    "maintainer": "Lennart Rolland",
    "maintainer_email": "lennart@firtkissdesigns.com",
    "description": ("CSS HTML JS SVG PNG JPEG Minifier"),
    "license": "GPL-3 LGPL-3 MIT",
    "platforms": ["Linux"],
    "keywords": "python3, CSS, HTML, JS, SVG, PNG, JPEG, Compressor, CSS3, HTML5, Web, Javascript, Minifier, Minify, Uglify, Obfuscator",
    "url": "https://gitlab.com/firstkiss/web-minifier",
    "download_url": "https://gitlab.com/firstkiss/web-minify",
    "project_urls": {"Docs": "https://gitlab.com/firstkiss/web-minify/README.md", "Bugs": "https://gitlab.com/firstkiss/web-minify/-/issues", "C.I.": "https://gitlab.com/firstkiss/web-minify/pipelines"},
    "packages": find_packages(),
    "zip_safe": True,
    "long_description": read_file("README.md"),
    "long_description_content_type": "text/markdown",
    "setup_requires": setup_requirements,
    "zip_safe": True,
    "install_requires": read_requirements_file("requirements/requirements.in"),  # Allow flexible deps for install
    "tests_require": read_requirements_file("requirements/test_requirements.txt"),  # Use rigid deps for testing
    "test_suite": "../tests",
    "python_requires": ">=3.7.4",
    "data_files": [("web-minify", [version_filename])],
    "include_package_data": True,
    # scripts=['css-html-js-minify.py'],  # uncomment if want install as script
    "entry_points": {"console_scripts": ["css-html-js-minify = css_html_js_minify.minify:main"]},
    "cmdclass": {"zipapp": ZipApp},
    # From https://pypi.org/pypi?%3Aaction=list_classifiers
    "classifiers": ["Development Status :: 3 - Alpha", "Intended Audience :: Developers", "Intended Audience :: Other Audience", "Topic :: Utilities", "Natural Language :: English", "Operating System :: POSIX :: Linux", "Programming Language :: Python :: 3.7", "Topic :: Other/Nonlisted Topic"],
}


pprint.pprint(package)
setup(**package)



import os
import pprint
import datetime
import time
from .handlers import html_minify, css_minify, js_minify, sass_minify, png_minify, jpeg_minify, svg_minify
from .files import path_is_in_path, generate_file_list, determine_file_extension
from .time import human_delta
import logging

logger = logging.getLogger(__name__)

now = datetime.datetime.now()


class Processor:
    def __init__(self, settings: dict):
        self.settings = settings
        self.input = settings.get("input", None)
        self.output = settings.get("output", None)
        self.input_is_dir = self.input and os.path.isdir(self.input)
        self.input_exists = self.input and os.path.exists(self.input)
        self.output_is_dir = self.output and os.path.isdir(self.output)
        self.output_exists = self.output and os.path.exists(self.output)

        self.overwrite: bool = settings.get("overwrite", False)
        self.on_change: bool = settings.get("on_change", False)
        self.format: bool = settings.get("format", False)
        self.prefix: bool = settings.get("prefix", False)
        self.zipy: bool = settings.get("zipy", False)
        self.hash: bool = settings.get("hash", False)
        self.verbose: bool = settings.get("verbose", False)
        self.force: bool = settings.get("force", False)

        # fmt off
        self.processor_map = {
            "sass": (sass_minify, False),
            "css": (css_minify, False),
            "html": (html_minify, False),
            "svg": (svg_minify, False),
            "js": (js_minify, False),
            "png": (png_minify, True),
            "jpeg": (jpeg_minify, True)
        }
        # fmt on

        self.valid_extensions = [f".{ext}" for ext in self.processor_map.keys()]
        if self.verbose:
            logger.info("### Settings:")
            logger.info(pprint.pformat(self.settings))
            logger.info("### Valid extensions:")
            logger.info(pprint.pformat(self.valid_extensions))
        self.pool = None

    def sanity_checks(self):
        if not self.input or not self.input_exists:
            return (False, f"The input specified '{self.input}' did not exist. Input must be an existing directory or file")
        if not self.overwrite and self.on_change:
            return (False, f"On-change will not have an effect so long as overwrite is not enabled")
        if not self.output and not self.overwrite and not self.format:
            return (False, f"Only input '{self.input}' was specified. Without a setting for 'output', 'overwrite' and/or 'format' all processing will fail")
        if path_is_in_path(self.input, self.output):
            return (False, f"The output '{self.output}' is a subpath of input '{self.input}'")
        if path_is_in_path(self.output, self.input):
            return (False, f"The input '{self.input}' is a subpath of output '{self.output}'")
        return True, None

    def process_file(self, input_path: str, output_path: str):
        if self.verbose:
            logger.info(f"Processing '{input_path}' to '{output_path}'")
        if not input_path:
            return False, False, False, "No input path specified"
        if not output_path:
            return False, False, False, "No output path specified"
        input_mtime = now.timestamp()
        if os.path.isfile(output_path):
            if not self.overwrite:
                return False, False, False, "Trying to overwrite without overwrite enabled"
            if self.on_change:
                input_mtime = os.path.getmtime(input_path)
                output_mtime = os.path.getmtime(output_path)
                if not self.force and (input_mtime == output_mtime):
                    return True, False, True, "File not changed, skipping"
                elif not self.force and (output_mtime > input_mtime):
                    return True, False, True, "Destination file newer than input, skipping"
        else:
            output_dir = os.path.dirname(output_path)
            if not os.path.isdir(output_dir):
                os.makedirs(name=output_dir, exist_ok=True)
            if not os.path.isdir(output_dir):
                return False, False, False, "Output directory did not exist or could not be created"
        extension = determine_file_extension(input_path)
        if not extension:
            return False, False, False, f"Unknown extension '{extension}' for input file"
        processor, is_binary = self.processor_map.get(extension, None)
        if not processor:
            return False, False, False, "Could not find processor for input file"
        original_content = None
        try:
            with open(input_path, "rb") if is_binary else open(input_path, "r", encoding="utf-8") as file:
                original_content = file.read()
        except Exception as err:
            return False, False, False, f"Could not load data from input file: {err}"
        if not original_content:
            logger.warning(f"Input file '{input_path}' was empty")
        processed_content = None
        was_copied = False
        if extension in ["js", "css", "html", "svg", "png", "jpeg"] and self.settings.get(f"disable_{extension}", False):
            # Perform copy
            # logger.info(f"SUPPOSED TO COPY: {input_path} ({extension})")
            processed_content = original_content
            was_copied = True
        else:
            # logger.info(f"SUPPOSED TO PROCESS: {input_path} ({extension})")
            processed_content = processor(original_content, self.settings)
            # TODO: make message mechanism for processor
        # logger.info(f"Content of file {input_path} was of type {type(original_content)} while processed {type(processed_content)}")
        if None == processed_content:
            logger.warning(f"Processed file for '{input_path}' was empty")
        try:
            with open(output_path, "wb") if is_binary else open(output_path, "w", encoding="utf-8") as file:
                written = file.write(processed_content)
                if written != len(processed_content):
                    return False, False, False, f"Partially written output ({written} of {len(processed_content)} bytes)"
        except Exception as err:
            return False, False, False, f"Could not write data to output file: {err}"
        try:
            os.utime(output_path, (input_mtime, input_mtime))
        except Exception as err:
            return False, False, False, f"Could not modify date of output file: {err}"
        # All went well, go home happy!
        return True, was_copied, False, None

    def process_files_list(self, list):
        start_time = datetime.datetime.now()
        success_count = 0
        copied_count = 0
        skipped_count = 0
        messages = {}
        # logger.info(f'Spreading workload over {multiprocessing.cpu_count()} CPU cores')
        # self.pool = multiprocessing.Pool(multiprocessing.cpu_count())
        for item in list:
            input_path = item["input_path"]
            output_path = item["output_path"]
            single_start_time = datetime.datetime.now()
            success, copied, skipped, message = self.process_file(input_path=input_path, output_path=output_path)
            if success:
                success_count += 1
            if copied:
                copied_count += 1
            if skipped:
                skipped_count += 1
            if message:
                messages[message] = 1 + messages.get(message, 0)

            single_end_time = datetime.datetime.now()
            single_interval = single_end_time - single_start_time
            if single_interval > datetime.timedelta(seconds=1):
                logger.warning(f"Processing of {input_path} was slow ({human_delta(single_interval)})")
        # self.pool.close()
        # self.pool.join()
        # self.pool=None
        end_time = datetime.datetime.now()
        interval = end_time - start_time
        failed = len(list) - success_count
        logger.info(f"Processing of {len(list)} files with {success_count} successful, {copied_count} copied and {skipped_count} skipped, generated {len(messages)} message(s) and took {human_delta(interval)} total")
        if messages:
            logger.warning(f"Messages encountered were:")
            for message, count in messages.items():
                logger.warning(f"{count} x {message}")
        return True

    def _process_existing_dir_to_existing_dir(self):
        input_paths = generate_file_list(self.input, tuple(self.valid_extensions))
        # logger.info(pprint.pformat(input_paths))
        list = []
        for input_path in input_paths:
            common = os.path.commonpath((os.path.abspath(self.output), os.path.abspath(input_path)))
            rel = os.path.relpath(os.path.abspath(input_path), os.path.abspath(self.input))
            output_path = os.path.join(os.path.abspath(self.output), rel)
            list.append({"input_path": input_path, "output_path": output_path})
        # logger.info(pprint.pformat(list))
        return list

    def _process_existing_dir_to_new_dir(self):
        # This is same as _process_existing_dir_to_existing_dir but with a mkdir first
        os.mkdir(self.output)
        self.output_exists = self.output and os.path.exists(self.output)
        return self._process_existing_dir_to_existing_dir()

    def process_files(self):
        if self.input_is_dir:
            if self.output:
                if self.output_is_dir:
                    if self.verbose:
                        logger.info(f"existing-dir-to-existing-dir easy peasy case")
                    list = self._process_existing_dir_to_existing_dir()
                    return self.process_files_list(list)
                elif not self.output_exists:
                    if self.verbose or True:
                        logger.info(f"existing-dir-to-new-dir easy peasy case")
                    list = self._process_existing_dir_to_new_dir()
                    return self.process_files_list(list)
                else:
                    logger.error(f"existing-dir-to-existing-file error")
                    return False
                return False
            elif self.overwite:
                return False
            elif self.format:
                return False
            else:
                logger.error(f"input dir specified without valid output option")
                return False
        else:
            if self.output:
                if self.output_is_dir:
                    if self.verbose or True:
                        logger.info(f"existing-file-to-existing-dir easy peasy case NOT IMPLEMENTED YET!")
                    return False
                elif not self.output_exists:
                    if self.verbose or True:
                        logger.info(f"existing-file-to-new-file easy peasy case NOT IMPLEMENTED YET!")
                    return False
                else:
                    if self.overwite:
                        if self.verbose or True:
                            logger.info(f"existing-file-to-existing-file overwrite easy peasy case NOT IMPLEMENTED YET!")
                        return False
                    else:
                        logger.error(f"existing-file-to-existing-file non-overwrite error")
                        return False
                return False
            elif self.overwite:
                return False
            elif self.format:
                return False
            else:
                logger.error(f"input file specified without valid output option")
                return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""HTML Minifier functions for CSS-HTML-JS-Minify."""


import re


__all__ = ("html_minify",)


def condense_html_whitespace(html):
    """Condense HTML, but be safe first if it have textareas or pre tags.

    >>> condense_html_whitespace('<i>  <b>    <a> test </a>    </b> </i><br>')
    '<i><b><a> test </a></b></i><br>'
    """  # first space between tags, then empty new lines and in-between.
    tagsStack = []
    split = re.split("(<\\s*pre.*>|<\\s*/\\s*pre\\s*>|<\\s*textarea.*>|<\\s*/\\s*textarea\\s*>)", html, flags=re.IGNORECASE)
    for i in range(0, len(split)):
        # if we are on a tag
        if (i + 1) % 2 == 0:
            tag = rawtag(split[i])
            if tag.startswith("/"):
                if not tagsStack or "/" + tagsStack.pop() != tag:
                    raise Exception("Some tag is not closed properly")
            else:
                tagsStack.append(tag)
            continue

        # else check if we are outside any nested <pre>/<textarea> tag
        if not tagsStack:
            temp = re.sub(r">\s+<", "> <", split[i])
            split[i] = re.sub(r"\s{2,}|[\r\n]", " ", temp)
    return "".join(split)


def rawtag(str):
    if re.match("<\\s*pre.*>", str, flags=re.IGNORECASE):
        return "pre"
    if re.match("<\\s*textarea.*>", str, flags=re.IGNORECASE):
        return "txt"
    if re.match("<\\s*/\\s*pre\\s*>", str, flags=re.IGNORECASE):
        return "/pre"
    if re.match("<\\s*/\\s*textarea\\s*>", str, flags=re.IGNORECASE):
        return "/txt"


def condense_style(html):
    """Condense style html tags.

    >>> condense_style('<style type="text/css">*{border:0}</style><p>a b c')
    '<style>*{border:0}</style><p>a b c'
    """  # May look silly but Emmet does this and is wrong.
    return html.replace('<style type="text/css">', "<style>").replace("<style type='text/css'>", "<style>").replace("<style type=text/css>", "<style>")


def condense_script(html):
    """Condense script html tags.

    >>> condense_script('<script type="text/javascript"> </script><p>a b c')
    '<script> </script><p>a b c'
    """  # May look silly but Emmet does this and is wrong.
    return html.replace('<script type="text/javascript">', "<script>").replace("<style type='text/javascript'>", "<script>").replace("<style type=text/javascript>", "<script>")


def clean_unneeded_html_tags(html):
    """Clean unneeded optional html tags.

    >>> clean_unneeded_html_tags('a<body></img></td>b</th></tr></hr></br>c')
    'abc'
    """
    for tag_to_remove in """</area> </base> <body> </body> </br> </col>
        </colgroup> </dd> </dt> <head> </head> </hr> <html> </html> </img>
        </input> </li> </link> </meta> </option> </param> <tbody> </tbody>
        </td> </tfoot> </th> </thead> </tr> </basefont> </isindex> </param>
            """.split():
        html = html.replace(tag_to_remove, "")
    return html  # May look silly but Emmet does this and is wrong.


def remove_html_comments(html):
    """Remove all HTML comments, Keep all for Grunt, Grymt and IE.

    >>> _="<!-- build:dev -->a<!-- endbuild -->b<!--[if IE 7]>c<![endif]--> "
    >>> _+= "<!-- kill me please -->keep" ; remove_html_comments(_)
    '<!-- build:dev -->a<!-- endbuild -->b<!--[if IE 7]>c<![endif]--> keep'
    """  # Grunt uses comments to as build arguments, bad practice but still.
    return re.compile(r"<!-- .*? -->", re.I).sub("", html)


def unquote_html_attributes(html):
    """Remove all HTML quotes on attibutes if possible.

    >>> unquote_html_attributes('<img   width="9" height="5" data-foo="0"  >')
    '<img width=9 height=5 data-foo=0 >'
    """  # data-foo=0> might cause errors on IE, we leave 1 space data-foo=0 >
    # cache all regular expressions on variables before we enter the for loop.
    any_tag = re.compile(r"<\w.*?>", re.I | re.MULTILINE | re.DOTALL)
    space = re.compile(r" \s+|\s +", re.MULTILINE)
    space1 = re.compile(r"\w\s+\w", re.MULTILINE)
    space2 = re.compile(r'"\s+>', re.MULTILINE)
    space3 = re.compile(r"'\s+>", re.MULTILINE)
    space4 = re.compile("\"\s\s+\w+=\"|'\s\s+\w+='|\"\s\s+\w+=|'\s\s+\w+=", re.MULTILINE)
    space6 = re.compile(r"\d\s+>", re.MULTILINE)
    quotes_in_tag = re.compile('([a-zA-Z]+)="([a-zA-Z0-9-_\.]+)"')
    # iterate on a for loop cleaning stuff up on the html markup.
    for tag in iter(any_tag.findall(html)):
        # exceptions of comments and closing tags
        if tag.startswith("<!") or tag.find("</") > -1:
            continue
        original = tag
        # remove white space inside the tag itself
        tag = space2.sub('" >', tag)  # preserve 1 white space is safer
        tag = space3.sub("' >", tag)
        for each in space1.findall(tag) + space6.findall(tag):
            tag = tag.replace(each, space.sub(" ", each))
        for each in space4.findall(tag):
            tag = tag.replace(each, each[0] + " " + each[1:].lstrip())
        # remove quotes on some attributes
        tag = quotes_in_tag.sub(r"\1=\2 ", tag)  # See Bug #28
        if original != tag:  # has the tag been improved ?
            html = html.replace(original, tag)
    return html.strip()


def add_timestamp(html):
    return f"{html}\n<!-- {format(datetime.now().isoformat()[:-7].lower())} -->\n"


def html_minify(html, settings: dict):
    """Minify HTML main function.

    >>> html_minify(' <p  width="9" height="5"  > <!-- a --> b </p> c <br> ')
    '<p width=9 height=5 > b c <br>'
    """
    comments = settings.get("comments", False)
    html = remove_html_comments(html) if not comments else html
    html = condense_style(html)
    html = condense_script(html)
    html = clean_unneeded_html_tags(html)
    html = condense_html_whitespace(html)
    html = unquote_html_attributes(html)
    html = add_timestamp(html) if settings.get("timestamp", False) else html
    return html.strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""JavaScript Minifier functions for CSS-HTML-JS-Minify."""


import re

from io import StringIO  # pure-Python StringIO supports unicode.

from .css import condense_semicolons


__all__ = ("js_minify",)


def remove_commented_lines(js):
    """Force remove commented out lines from Javascript."""
    result = ""
    for line in js.splitlines():
        line = re.sub(r"/\*.*\*/", "", line)  # (/*COMMENT */)
        line = re.sub(r"//.*", "", line)  # (//COMMENT)
        result += "\n" + line
    return result


def simple_replacer_js(js):
    """Force strip simple replacements from Javascript."""
    return condense_semicolons(js.replace("debugger;", ";").replace(";}", "}").replace("; ", ";").replace(" ;", ";").rstrip("\n;"))


def js_minify_keep_comments(js):
    """Return a minified version of the Javascript string."""
    ins, outs = StringIO(js), StringIO()
    JavascriptMinify(ins, outs).minify()
    return force_single_line_js(outs.getvalue())


def force_single_line_js(js):
    """Force Javascript to a single line, even if need to add semicolon."""
    return ";".join(js.splitlines()) if len(js.splitlines()) > 1 else js


class JavascriptMinify(object):

    """Minify an input stream of Javascript, writing to an output stream."""

    def __init__(self, instream=None, outstream=None):
        """Init class."""
        self.ins, self.outs = instream, outstream

    def minify(self, instream=None, outstream=None):
        """Minify Javascript using StringIO."""
        if instream and outstream:
            self.ins, self.outs = instream, outstream
        write, read = self.outs.write, self.ins.read
        space_strings = "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_$\\"
        starters, enders = "{[(+-", "}])+-\"'"
        newlinestart_strings = starters + space_strings
        newlineend_strings = enders + space_strings
        do_newline, do_space = False, False
        doing_single_comment, doing_multi_comment = False, False
        previous_before_comment, in_quote = "", ""
        in_re, quote_buf = False, []
        previous = read(1)
        next1 = read(1)
        if previous == "/":
            if next1 == "/":
                doing_single_comment = True
            elif next1 == "*":
                doing_multi_comment = True
            else:
                write(previous)
        elif not previous:
            return
        elif previous >= "!":
            if previous in "'\"":
                in_quote = previous
            write(previous)
            previous_non_space = previous
        else:
            previous_non_space = " "
        if not next1:
            return
        while True:
            next2 = read(1)
            if not next2:
                last = next1.strip()
                conditional_1 = doing_single_comment or doing_multi_comment
                if not conditional_1 and last not in ("", "/"):
                    write(last)
                break
            if doing_multi_comment:
                if next1 == "*" and next2 == "/":
                    doing_multi_comment = False
                    next2 = read(1)
            elif doing_single_comment:
                if next1 in "\r\n":
                    doing_single_comment = False
                    while next2 in "\r\n":
                        next2 = read(1)
                        if not next2:
                            break
                    if previous_before_comment in ")}]":
                        do_newline = True
                    elif previous_before_comment in space_strings:
                        write("\n")
            elif in_quote:
                quote_buf.append(next1)

                if next1 == in_quote:
                    numslashes = 0
                    for c in reversed(quote_buf[:-1]):
                        if c != "\\":
                            break
                        else:
                            numslashes += 1
                    if numslashes % 2 == 0:
                        in_quote = ""
                        write("".join(quote_buf))
            elif next1 in "\r\n":
                conditional_2 = previous_non_space in newlineend_strings
                if conditional_2 or previous_non_space > "~":
                    while 1:
                        if next2 < "!":
                            next2 = read(1)
                            if not next2:
                                break
                        else:
                            conditional_3 = next2 in newlinestart_strings
                            if conditional_3 or next2 > "~" or next2 == "/":
                                do_newline = True
                            break
            elif next1 < "!" and not in_re:
                conditional_4 = next2 in space_strings or next2 > "~"
                conditional_5 = previous_non_space in space_strings
                conditional_6 = previous_non_space > "~"
                if (conditional_5 or conditional_6) and (conditional_4):
                    do_space = True
            elif next1 == "/":
                if in_re:
                    if previous != "\\":
                        in_re = False
                    write("/")
                elif next2 == "/":
                    doing_single_comment = True
                    previous_before_comment = previous_non_space
                elif next2 == "*":
                    doing_multi_comment = True
                else:
                    in_re = previous_non_space in "(,=:[?!&|"
                    write("/")
            else:
                if do_space:
                    do_space = False
                    write(" ")
                if do_newline:
                    write("\n")
                    do_newline = False
                write(next1)
                if not in_re and next1 in "'\"":
                    in_quote = next1
                    quote_buf = []
            previous = next1
            next1 = next2
            if previous >= "!":
                previous_non_space = previous


def add_timestamp(js):
    return f"{js}\n/* {format(datetime.now().isoformat()[:-7].lower())} */\n"


def js_minify(js, settings: dict):
    """Minify a JavaScript string."""
    """Future JavaScript support is orphan and not supported!.
       If you want to make ES6,ES7 work feel free to send pull requests.
    """
    if not settings.get("comments", False):
        js = remove_commented_lines(js)
    js = js_minify_keep_comments(js)
    js = add_timestamp(js) if settings.get("timestamp", False) else js
    return js.strip()
